from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Callable, Generator, Protocol, TypeVar, overload

from typing_extensions import Never

from ucam_faas.exceptions import with_notes

if TYPE_CHECKING:
    from contextlib import _GeneratorContextManager

_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")


class ActivateContextVarFn(Protocol[_T]):
    """Get a context manager that sets `value` on a `ContextVar` while active."""

    def __call__(self, value: _T | None = None) -> _GeneratorContextManager[_T]:
        pass


def contextvar_activator(contextvar: ContextVar[_T]) -> ActivateContextVarFn[_T]:
    """
    Create a function that sets a `ContextVar`'s value within a `with` block.

    When called with an existing `ContextVar`, this function returns another
    function that itself takes a value to set on the `ContextVar` and returns a
    context manager that will activate and deactivate the value on the
    `ContextVar`.

    Example
    -------
    >>> my_context = ContextVar('my_context', default='Not Set')
    >>> activate_my_context = contextvar_activator(my_context)

    >>> my_context.get()
    'Not Set'

    >>> with activate_my_context('Hello World!') as active_value:
    ...     assert active_value == 'Hello World!'

    >>> my_context.get()
    'Not Set'
    """

    @contextmanager
    def activate_context(value: _T | None = None) -> Generator[_T]:
        token: Token[_T] | None = None
        if value is not None:
            token = contextvar.set(value)
        try:
            yield contextvar.get()
        finally:
            if token is not None:
                contextvar.reset(token)

    activate_context.__name__ = f"activate_{contextvar.name.lower()}"
    return activate_context


@overload
def contextual_default_factory(
    contextvar: ContextVar[_T],
    *,
    fn: None = None,
    default_fn: None = None,
) -> Callable[[], _T]:
    ...


@overload
def contextual_default_factory(
    contextvar: ContextVar[_T],
    *,
    fn: Callable[[_T], _U],
    default_fn: Callable[[LookupError], Never] | None = None,
) -> Callable[[], _U]:
    ...


@overload
def contextual_default_factory(
    contextvar: ContextVar[_T],
    *,
    fn: Callable[[_T], _U],
    default_fn: Callable[[LookupError], _V],
) -> Callable[[], _U | _V]:
    ...


@overload
def contextual_default_factory(
    contextvar: ContextVar[_T],
    *,
    default_fn: Callable[[LookupError], _V],
    fn: None = None,
) -> Callable[[], _T | _V]:
    ...


def contextual_default_factory(
    contextvar: ContextVar[_T],
    *,
    fn: Callable[[_T], _U] | None = None,
    default_fn: Callable[[LookupError], _V] | None = None,
) -> Callable[[], _T | _U | _V]:
    """
    Create a default-providing function that refers to the value of a `ContextVar`.

    When the returned function is called to provide a default, the `fn`
    keyword argument is called with the current value of the `contextvar` and it
    returns the default value to use. If `fn` is not provided, the `contextvar`'s
    value is returned directly as the default.

    When the `contextvar` holds no value, the `default_fn` is called with a
    `LookupError`. It can either return a default value or raise an error. If
    `default_fn` is not provided, the `LookupError` is raised.
    """

    def contextual_default_factory() -> _T | _U | _V:
        try:
            ctx = contextvar.get()
        except LookupError as e:
            e = with_notes(
                e,
                "A ContextVar was empty when its value was read by a "
                "default_factory function to provide a default value. "
                "This is often caused by trying to create an instance of a "
                "Pydantic model or dataclass that gets default values for its "
                "fields from the environment it's created in. You either need "
                "to provide an explicit value for the field(s) that use a "
                "contextual_default_factory(), or create it in an environment "
                "that provides a value for this empty ContextVar.",
            )
            if default_fn is not None:
                return default_fn(e)
            raise e
        return ctx if fn is None else fn(ctx)

    return contextual_default_factory



"""
Support for parsing/validating/serialising GCP Pub/Sub messages within
CloudEvent envelopes.

This module provides constants for the CloudEvent type that identifies message
events, and Pydantic data models for the data payload of these events.

Example
-------
>>> cloud_event_json = {
...     "id": "1",
...     "time": "2025-03-18T09:01:02.345Z",
...     "specversion": "1.0",
...     "datacontenttype": "application/json",
...     "type": "google.cloud.pubsub.topic.v1.messagePublished",
...     "source": "//pubsub.googleapis.com/projects/my-project/topics/my-topic",
...     "data": {
...         "message": {
...             "@type": "type.googleapis.com/google.pubsub.v1.PubsubMessage",
...             "data": "bWVzc2FnZTE=",
...             "attributes": {"my_attr_1": "foo", "my_attr_2": "bar"},
...             "messageId": "1",
...             "publishTime": "2025-03-18T09:01:02.345Z",
...         }
...     },
... }
>>> from cloudevents.conversion import from_dict
>>> from cloudevents.pydantic.v2 import CloudEvent as PydanticCloudEvent
>>> cloud_event = from_dict(PydanticCloudEvent, cloud_event_json)
>>>
>>> assert cloud_event.get_attributes()['type'] == PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE
>>>
>>> payload = MessagePublishedData.model_validate(cloud_event.get_data())
>>> assert payload.message.data == b'message1'
>>> assert payload.message.message_id == '1'
>>> assert payload.message.attributes == {"my_attr_1": "foo", "my_attr_2": "bar"}
>>> assert payload.message.publish_time.year == 2025
>>>
>>> # The Pub/Sub topic is in the CloudEvent's "source" attribute.
>>> assert cloud_event.get_attributes()['source'] == \\
...     "//pubsub.googleapis.com/projects/my-project/topics/my-topic"
>>>
>>> # Although we can parse the project_id and topic_id as follows, if you need
>>> # to verify the topic is correct, it's probably fine to treat the source as
>>> # an atomic/opaque value and compare it to a string containing your IDs.
>>> topic = cloud_event.get_attributes()['source'].removeprefix('//pubsub.googleapis.com/')
>>> PUBSUB_CLOUD_EVENT_TOPIC_PATTERN.match(topic).groupdict()
{'project_id': 'my-project', 'topic_id': 'my-topic'}
"""
from __future__ import annotations

import base64
import re
from typing import TYPE_CHECKING, Annotated, Final, Literal, Mapping

from pydantic import (
    AwareDatetime,
    BaseModel,
    ConfigDict,
    Field,
    PlainSerializer,
    PlainValidator,
    StringConstraints,
)
from pydantic_core import PydanticCustomError

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

PubsubMessageType: TypeAlias = Literal["type.googleapis.com/google.pubsub.v1.PubsubMessage"]
PUBSUB_MESSAGE_TYPE: Final = "type.googleapis.com/google.pubsub.v1.PubsubMessage"
"""The `@type` property of Pubsub message data within `...messagePublished` CloudEvent objects."""
PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE: Final = "google.cloud.pubsub.topic.v1.messagePublished"
"""The `type` property used on CloudEvent objects containing `...PubsubMessage` data."""

PUBSUB_CLOUD_EVENT_TOPIC_PATTERN: Final = re.compile(
    r"^projects/(?P<project_id>[^/]+)/topics/(?P<topic_id>[^/]+)$"
)

NonEmptyString: TypeAlias = Annotated[str, StringConstraints(min_length=1)]


def _decode_Base64JsonString(data: str | bytes) -> bytes:
    if isinstance(data, bytes):
        return data
    try:
        return base64.b64decode(data)
    except ValueError as e:
        raise PydanticCustomError(
            "base64_decode", "Base64 decoding error: '{error}'", {"error": str(e)}
        )


def _encode_Base64JsonString(data: bytes) -> str:
    return base64.b64encode(data).decode()


Base64JsonString: TypeAlias = Annotated[
    bytes,
    PlainSerializer(func=_encode_Base64JsonString, when_used="json"),
    PlainValidator(func=_decode_Base64JsonString, json_schema_input_type=str),
]
"""Binary data. Python representation is bytes, JSON is a (non-URL-safe) base64 string."""


class PubsubMessage(BaseModel):
    """
    A message from Pubsub to notify a subscriber of an event on their topic.

    The message data is a byte string in the application-specific format known to
    the subscriber.
    """

    model_config = ConfigDict(populate_by_name=True)

    type: PubsubMessageType = Field(alias="@type", default=PUBSUB_MESSAGE_TYPE)
    data: Base64JsonString
    attributes: Mapping[str, str]
    message_id: NonEmptyString = Field(alias="messageId")
    publish_time: AwareDatetime = Field(alias="publishTime")


class MessagePublishedData(BaseModel):
    """
    The Pubsub `MessagePublishedData` object.

    This JSON object wraps a `PubsubMessage` and in theory could contain extra
    information about it. In practice, it contains only a message property and
    no other fields, but the schema suggests it could contain `subscription` and
    `deliveryAttempt` properties. When the `functions_framework` constructs it,
    it never populates these, only `message`.

    See Also
    --------
    - https://github.com/GoogleCloudPlatform/functions-framework-conformance/\
blob/main/docs/mapping.md#cloud-pubsub-events
    - [google-cloudevents JSON schema](https://github.com/googleapis/\
google-cloudevents/blob/main/jsonschema/google/events/cloud/pubsub/v1/\
MessagePublishedData.json)
    """

    message: PubsubMessage



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest
import structlog
from cloudevents import conversion
from flask import Flask
from structlog.testing import capture_logs

from ucam_faas import FaaSGunicornApplication, cloud_event, raw_event, run_ucam_faas
from ucam_faas.exceptions import UCAMFAASCouldNotProcess
from ucam_faas.testing import CreateEventAppClientFn

if TYPE_CHECKING:
    from typing_extensions import Never


script_path = Path(__file__).resolve()
cwd = Path.cwd()
relative_path = script_path.relative_to(cwd)


# Raw events
@raw_event
def example_raw_event_no_exception(event: bytes) -> None:
    pass


@raw_event
def example_raw_event_handled_exception(event: bytes) -> Never:
    raise UCAMFAASCouldNotProcess


@raw_event
def example_raw_event_unhandled_exception(event: bytes) -> Never:
    raise Exception("Did not expect this")


# Cloud Events
@cloud_event
def example_cloud_event_no_exception(event: Any) -> None:
    pass


@cloud_event
def example_cloud_event_handled_exception(event: Any) -> None:
    raise UCAMFAASCouldNotProcess


@cloud_event
def example_cloud_event_unhandled_exception(event: Any) -> None:
    raise Exception("Did not expect this")


def test_faas_gunicorn_application_bind() -> None:
    app = Flask(__name__)
    application = FaaSGunicornApplication(app, "0.0.0.0", "8080")

    with patch("gunicorn.app.base.BaseApplication.run") as mock_run:
        application.run()
        mock_run.assert_called_once()  # Ensures that the server's run method was indeed called


@pytest.mark.parametrize(
    "target_tuple",
    [
        (
            "example_raw_event_no_exception",
            "example_raw_event_handled_exception",
            "example_raw_event_unhandled_exception",
        ),
        (
            "example_cloud_event_no_exception",
            "example_cloud_event_handled_exception",
            "example_cloud_event_unhandled_exception",
        ),
    ],
)
def test_exceptions_raw_events(
    event_app_test_client_factory: CreateEventAppClientFn,
    target_tuple: tuple[str, str, str],
    valid_cloud_event: Any,
) -> None:
    # No exception
    test_client = event_app_test_client_factory(target=target_tuple[0], source=relative_path)
    response = test_client.post("/", json=conversion.to_dict(valid_cloud_event))
    assert response.status_code == 200

    # Handle exception
    test_client = event_app_test_client_factory(target=target_tuple[1], source=relative_path)
    response = test_client.post("/", json=conversion.to_dict(valid_cloud_event))
    assert response.status_code == 500
    assert "The function raised UCAMFAASCouldNotProcess" in response.data.decode()

    # Unhandled exception
    test_client = event_app_test_client_factory(target=target_tuple[2], source=relative_path)
    with structlog.testing.capture_logs() as cap_logs:
        response = test_client.post("/", json=conversion.to_dict(valid_cloud_event))
    assert response.status_code == 500

    assert len(cap_logs) == 1
    log_call = cap_logs[0]
    assert (
        log_call.get("event") == "function_failed_uncaught_exception"
        # structlog will include exception details in rendered log message
        and log_call.get("exc_info") is True
    )


@pytest.mark.parametrize(
    "target,expected_return",
    [
        ("example_raw_event_no_exception", 0),
        ("example_raw_event_handled_exception", 1),
        ("example_raw_event_unhandled_exception", 2),
        ("example_cloud_event_no_exception", 0),
        ("example_cloud_event_handled_exception", 1),
        ("example_cloud_event_unhandled_exception", 2),
    ],
)
def test_long_running(target: str, expected_return: int) -> None:
    assert (
        run_ucam_faas(
            target,
            source=relative_path,
            host="no-host",
            port=0,
            debug=False,
            long_running=True,
        )
        == expected_return
    )


def normal_function() -> None:
    pass


def test_long_running_no_registered_event_handler() -> None:
    """
    Test that running a normal function as a long-running function raises an error.
    """
    with capture_logs() as cap_logs:
        assert (
            run_ucam_faas(
                normal_function.__name__,
                source=relative_path,
                host="no-host",
                port=0,
                debug=False,
                long_running=True,
            )
            == 2
        )

    assert len(cap_logs) == 1
    assert cap_logs[0]["event"] == "function_handler_not_registered"



from dataclasses import dataclass

from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    CompletedExecutionResult,
    ExecutionInfo,
    ExecutionStatus,
)


class TestCompletedExecutionResult:
    def test_init(self) -> None:
        result = CompletedExecutionResult()
        assert result.status is ExecutionStatus.COMPLETED
        assert result.duration_seconds is None

        result = CompletedExecutionResult(duration_seconds=1.5)
        assert result.duration_seconds == 1.5


class TestAbortedExecutionResult:
    def test_init(self) -> None:
        err = ValueError("Oops")
        result = AbortedExecutionResult(
            reason=AbortedExecutionReason.USER_DATA_INVALID,
            exception=err,
            duration_seconds=1.5,
        )
        assert result.status is ExecutionStatus.ABORTED
        assert result.reason is AbortedExecutionReason.USER_DATA_INVALID
        assert result.exception is err
        assert result.duration_seconds == 1.5

        result = AbortedExecutionResult(reason="custom.reason")
        assert result.status is ExecutionStatus.ABORTED
        assert result.reason == "custom.reason"
        assert result.exception is None
        assert result.duration_seconds is None


class TestExecutionInfo:
    @dataclass
    class CustomAttributes:
        size: int
        label: str

    def test_init(self) -> None:
        execinfo = ExecutionInfo()
        assert execinfo.execution == CompletedExecutionResult()

        execinfo = ExecutionInfo(
            execution=AbortedExecutionResult(
                reason=AbortedExecutionReason.USER_DATA_INVALID, exception=ValueError("Oops")
            )
        )
        assert execinfo.execution.status is ExecutionStatus.ABORTED
        assert isinstance(execinfo.execution.exception, ValueError)

    def test_arbitrary_attributes(self) -> None:
        """ExecutionInfo models can hold arbitrary attributes without subclassing."""
        custom = TestExecutionInfo.CustomAttributes(size=42, label="foo")
        execinfo = ExecutionInfo(custom_stuff=custom)

        assert execinfo.custom_stuff == custom  # type: ignore[attr-defined]
        assert execinfo.model_extra is not None
        assert execinfo.model_extra["custom_stuff"] == custom

        assert execinfo.model_dump(mode="json") == {
            "execution": {"duration_seconds": None, "status": "COMPLETED"},
            "custom_stuff": {"label": "foo", "size": 42},
        }

    def test_from_context(self) -> None:
        assert ExecutionInfo.from_context(execution=CompletedExecutionResult()) == ExecutionInfo(
            execution=CompletedExecutionResult()
        )



from __future__ import annotations

from traceback import TracebackException
from typing import TYPE_CHECKING, Annotated, Union

from pydantic import PlainSerializer

if TYPE_CHECKING:
    from typing_extensions import TypeAlias


def _serialize_exception(value: Exception | str) -> str:
    if isinstance(value, str):
        return value
    return "".join(TracebackException.from_exception(value).format())


SerialisableException: TypeAlias = Annotated[
    Union[Exception, str], PlainSerializer(_serialize_exception, when_used="json")
]
"""A Pydantic type for Exceptions that are JSON-serialised as a traceback string."""



"""
Support for event handler functions that process CloudEvents.

This module provides support for:

- Validating the type of incoming CloudEvents
- Parsing/validating the data payload of CloudEvents
- Recording the CloudEvent's metadata in the function's ExecutionInfo result

Handler functions can focus on processing a primary data value without needing
to do their own parsing and validation or event metadata logging/reporting.

To use this module, handler functions are wrapped with this module's
`@cloud_event_handler(...)` decorator. The decorator takes a parser function
that is responsible for validating CloudEvent values, and extracting their data
payload.

Handler functions receive the primary value from the parser function, and also
have access to the the full CloudEvent via the `CLOUD_EVENT_HANDLER_CONTEXT`
`ContextVar`. Metadata from the CloudEvent is automatically included in
`CloudEventHandlerExecutionInfo` values created within the handler function's
scope.
"""

from __future__ import annotations

import functools
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Protocol, cast, overload

from cloudevents.abstract import CloudEvent
from pydantic import BaseModel, Field, ValidationError
from typing_extensions import Never, Self, TypeVar

from ucam_faas.contexts import contextual_default_factory, contextvar_activator
from ucam_faas.contextual_handlers import (
    ContextualHandlerFn,
    ContextualHandlerParserFn,
    CreateExecutionInfoFn,
    HandlerFn,
    contextual_handler,
)
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    ExecutionInfo,
)
from ucam_faas.handler_execution import (
    DEFAULT_EXECUTION_RESULT_LOGGER,
    DEFAULT_HTTP_RESPONSE_CREATOR,
    CloudEventFunctionsFrameworkHandlerExecutorFn,
    HandlerExecutionOptions,
    cloud_event_handler_executor,
    get_function_execution_id,
)

CLOUD_EVENT_CONTEXT: ContextVar[CloudEventHandlerContext] = ContextVar(
    f"{__name__}.CLOUD_EVENT_HANDLER_CONTEXT"
)
activate_cloud_event_context = contextvar_activator(CLOUD_EVENT_CONTEXT)

_PydanticModelT = TypeVar("_PydanticModelT", bound=BaseModel)
_T = TypeVar("_T")
_T_con = TypeVar("_T_con", contravariant=True)
_T_co = TypeVar("_T_co", covariant=True)

_InfoT = TypeVar("_InfoT", bound=ExecutionInfo)
_InfoT_co = TypeVar("_InfoT_co", bound=ExecutionInfo, covariant=True)
_DefaultInfoT = TypeVar("_DefaultInfoT", bound=ExecutionInfo)
_DefaultInfoT_co = TypeVar("_DefaultInfoT_co", bound=ExecutionInfo, covariant=True)
_UsedDefaultInfoT_co = TypeVar(
    "_UsedDefaultInfoT_co", bound=ExecutionInfo, covariant=True, default=Never
)

_NoneT = TypeVar("_NoneT", None, Never, default=Never)
_NoneT_co = TypeVar("_NoneT_co", None, Never, covariant=True, default=Never)


@dataclass
class CloudEventHandlerContext:
    cloud_event: CloudEvent
    """The CloudEvent that is being handled."""
    execution_id: str | None
    """The unique identifier for the currently-executing GCP Cloud Run function."""

    handler: CloudEventHandlerAttributes
    """Metadata on the active `@cloud_event_handler`."""


@dataclass
class CloudEventAttributes:
    """
    Attribute metadata from a CloudEvent.

    The fields are named to follow OpenTelemetry Semantic Conventions for
    CloudEvents: https://opentelemetry.io/docs/specs/semconv/attributes-registry/cloudevents/
    """

    event_id: str
    event_source: str
    event_spec_version: str
    event_subject: str | None
    event_type: str
    event_time: str | None  # not currently in opentelemetry semconv

    @classmethod
    def from_cloud_event(cls, cloud_event: CloudEvent) -> Self:
        attributes = cloud_event.get_attributes()
        return cls(
            # Fields accessed without .get() are mandatory
            event_id=attributes["id"],
            event_source=attributes["source"],
            event_spec_version=attributes["specversion"],
            event_subject=attributes.get("subject") or None,
            event_type=attributes["type"],
            event_time=attributes.get("time") or None,
        )


@dataclass
class CloudEventHandlerAttributes:
    input_event_accepted: bool


class CloudEventHandlerExecutionInfo(ExecutionInfo):
    execution_id: str | None = Field(
        default_factory=contextual_default_factory(
            CLOUD_EVENT_CONTEXT, fn=lambda ctx: ctx.execution_id
        )
    )
    cloudevents: CloudEventAttributes = Field(
        default_factory=contextual_default_factory(
            CLOUD_EVENT_CONTEXT,
            fn=lambda ctx: CloudEventAttributes.from_cloud_event(ctx.cloud_event),
        )
    )
    cloud_event_handler: CloudEventHandlerAttributes = Field(
        default_factory=contextual_default_factory(
            CLOUD_EVENT_CONTEXT,
            fn=lambda ctx: ctx.handler,
        )
    )


class CloudEventDataParserFn(Protocol[_T_co]):
    def __call__(self, cloud_event: CloudEvent, /) -> _T_co | ExecutionInfo:
        pass


@dataclass
class CloudEventType(CloudEventDataParserFn[_PydanticModelT]):
    """An `event_data_parser` function for `@cloud_event_handler`.

    The function validates that a CloudEvent has a specific type, and validates
    the event's data by parsing it with a Pydantic model.
    """

    event_type: str
    event_data_model: type[_PydanticModelT]

    def validate_cloud_event(
        self, cloud_event: CloudEvent
    ) -> _PydanticModelT | CloudEventHandlerExecutionInfo:
        try:
            if (t := cloud_event.get_attributes().get("type")) != self.event_type:
                raise ValueError(f"Received CloudEvent with incorrect type: {t!r}")
        except ValueError as e:
            return CloudEventHandlerExecutionInfo(
                execution=AbortedExecutionResult(
                    reason=AbortedExecutionReason.CLOUD_EVENT_TYPE_INCORRECT,
                    exception=e,
                )
            )

        try:
            ce_data = self.event_data_model.model_validate(cloud_event.get_data())
        except ValidationError as e:
            return CloudEventHandlerExecutionInfo(
                execution=AbortedExecutionResult(
                    reason=AbortedExecutionReason.SERVICE_DATA_INVALID, exception=e
                )
            )

        return ce_data

    def __call__(
        self, cloud_event: CloudEvent, /
    ) -> _PydanticModelT | CloudEventHandlerExecutionInfo:
        return self.validate_cloud_event(cloud_event)


# Used as a TypeAlias to shorten the type name of @cloud_event_handler.
class CloudEventHandlerFn(
    ContextualHandlerFn[CloudEvent, _T_con, _InfoT_co, _UsedDefaultInfoT_co, _NoneT_co],
    Protocol[_T_con, _InfoT_co, _UsedDefaultInfoT_co, _NoneT_co],
):
    """The type of the inner handler function created by `@cloud_event_handler`."""


class CloudEventHandlerExecutorFn(
    CloudEventFunctionsFrameworkHandlerExecutorFn[
        CloudEventHandlerFn[_T_con, _InfoT_co, _UsedDefaultInfoT_co, _NoneT_co]
    ],
    Protocol[_T_con, _InfoT_co, _UsedDefaultInfoT_co, _NoneT_co],
):
    """
    The type of the outer `functions_framework` wrapper created by `@cloud_event_handler`.

    The `execute_handler()` method is the handler function that returns `ExecutionInfo`.
    """


class CloudEventHandlerDecoratorFn(Protocol[_T, _DefaultInfoT_co]):
    """The type of the decorator returned by `@cloud_event_handler(...)`"""

    # These overloads are needed to correctly type the return value of
    # the decorated handler function. The return varies according to whether the
    # handler function returns None:
    #
    # - When the handler CAN return None, the return is either the default
    #   ExecutionInfo type or the handler's non-None return type (if any).
    #   returned.
    # - When the handler NEVER returns None, the return is always the
    #   handler's own return type.

    @overload
    def __call__(
        self, event_data_handler: HandlerFn[_T, _InfoT], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT]:
        pass

    @overload
    def __call__(
        self, event_data_handler: HandlerFn[_T, _InfoT | None], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT, _DefaultInfoT_co, None]:
        pass

    def __call__(
        self, event_data_handler: HandlerFn[_T, _InfoT | _NoneT], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT, _DefaultInfoT_co, _NoneT]:
        pass


@overload
def cloud_event_handler(
    *,
    event_data_parser: CloudEventDataParserFn[_T],
    create_default_info: None = None,
    options: HandlerExecutionOptions | None = None,
) -> CloudEventHandlerDecoratorFn[_T, CloudEventHandlerExecutionInfo]:
    pass


@overload
def cloud_event_handler(
    *,
    event_data_parser: CloudEventDataParserFn[_T],
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT],
    options: HandlerExecutionOptions | None = None,
) -> CloudEventHandlerDecoratorFn[_T, _DefaultInfoT]:
    pass


def cloud_event_handler(
    *,
    event_data_parser: CloudEventDataParserFn[_T],
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT] | None = None,
    options: HandlerExecutionOptions | None = None,
) -> CloudEventHandlerDecoratorFn[_T, _DefaultInfoT]:
    """
    Register a function to be called when a CloudEvent needs to be handled.

    Parameters
    ----------
    event_data_parser:
        A function used to parse and validate CloudEvents before calling the
        decorated handler function. The returned value is the value the handler
        function receives.

        If the CloudEvent is not valid (wrong type, invalid data, etc), the
        function must provide an `ExecutionInfo` result instead of a value,
        either by returning it or throwing `ExecutionAbortedException`.
    create_default_info:
        A function used to create an ExecutionInfo value if the decorated
        handler function does not return one.
    options:
        Override the default behaviour for logging the `ExecutionInfo` result,
        and representing the `ExecutionInfo` result as an HTTP response.

    Example
    -------
    First define the data payload of the event you will handle:
    >>> from pydantic import BaseModel
    >>> class GreetingEventData(BaseModel):
    ...     greeting: str

    Second, define the CloudEvent that the data payload is delivered within:
    >>> from ucam_faas import CloudEventType
    >>> parse_greeting_event = CloudEventType(
    ...     event_type="com.example.events.greeting",
    ...     event_data_model=GreetingEventData,
    ... )

    Third, define a function to handle the event. It'll be called automatically
    each time an event is received — as long as the event passes validation
    using your model and event type.
    >>> from ucam_faas import cloud_event_handler
    >>> @cloud_event_handler(event_data_parser=parse_greeting_event)
    ... def handle_greeting(event_data: GreetingEventData):
    ...     print(f"Received greeting: {event_data.greeting!r}")

    The handler function can be executed for testing purposes:
    >>> from ucam_faas.tests.fixtures import make_cloud_event
    >>> exc_info = handle_greeting.execute_for_result(
    ...     make_cloud_event(
    ...         type="com.example.events.greeting",
    ...         source="testing",
    ...         data={"greeting": "Hello World!"}
    ...     ),
    ... )
    Received greeting: 'Hello World!'
    """  # noqa: E501

    @overload
    def cloud_event_handler_decorator(
        event_data_handler: HandlerFn[_T, _InfoT], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT]:
        pass

    @overload
    def cloud_event_handler_decorator(
        event_data_handler: HandlerFn[_T, _InfoT | None], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT, _DefaultInfoT, None]:
        pass

    def cloud_event_handler_decorator(
        event_data_handler: HandlerFn[_T, _InfoT | _NoneT], /
    ) -> CloudEventHandlerExecutorFn[_T, _InfoT, _DefaultInfoT, _NoneT]:
        _options = options or HandlerExecutionOptions()

        # When create_default_info is None or
        # CreateExecutionInfoFn[CloudEventHandlerExecutionInfo],
        # _DefaultInfoT IS CloudEventHandlerExecutionInfo because of the
        # @overloads, so this cast is safe.
        _create_default_info = cast(
            CreateExecutionInfoFn[_DefaultInfoT],
            create_default_info or CloudEventHandlerExecutionInfo.from_context,
        )

        # Wrap our augmented handler to make it suitable for being called by
        # functions_framework as a CloudEvent handler. This deals with logging
        # the ExecutionInfo result, and sending a suitable HTTP response,
        # according to the result.
        @cloud_event_handler_executor(
            create_http_response=_options.make_http_response or DEFAULT_HTTP_RESPONSE_CREATOR,
            log_execution_result=_options.log_execution_result or DEFAULT_EXECUTION_RESULT_LOGGER,
        )
        # Wrap the raw handler function in order to parse the raw CloudEvent
        # to extract the handler's input value, as well as to provide the
        # CloudEvent metadata in the environment via a ContextVar. Also to
        # create a default ExecutionResult value if the handler throws or
        # returns None.
        @contextual_handler(
            parser=contextual_cloud_event_parser(event_data_parser),
            context=activate_cloud_event_context,
            default_info_factory=_create_default_info,
        )
        @functools.wraps(event_data_handler)
        def event_data_handler_executor(event_data: _T, /) -> _InfoT | _NoneT:
            return event_data_handler(event_data)

        assert event_data_handler_executor.__name__ == event_data_handler.__name__
        return event_data_handler_executor

    return cloud_event_handler_decorator


def contextual_cloud_event_parser(
    cloud_event_data_parser: CloudEventDataParserFn[_T_co],
) -> ContextualHandlerParserFn[CloudEvent, _T_co, CloudEventHandlerContext, ExecutionInfo]:
    """Wrap a (CloudEvent) -> T function to be a @contextual_handler() parser.

    The T value returned by the function is the primary value and the CloudEvent
    is in the context value's `cloud_event` attribute.

    The function is run with `CLOUD_EVENT_CONTEXT` active.
    """

    @functools.wraps(cloud_event_data_parser)
    def parse_input(
        cloud_event: CloudEvent, /
    ) -> tuple[_T_co, CloudEventHandlerContext] | ExecutionInfo:
        context = CloudEventHandlerContext(
            cloud_event=cloud_event,
            execution_id=get_function_execution_id(),
            handler=CloudEventHandlerAttributes(input_event_accepted=False),
        )

        with activate_cloud_event_context(context):
            parse_result = cloud_event_data_parser(cloud_event)

        if isinstance(parse_result, ExecutionInfo):
            return parse_result

        context.handler.input_event_accepted = True
        return parse_result, context

    return parse_input



from __future__ import annotations

import http.client
import types
from typing import Final, Generator, Protocol
from unittest import mock
from unittest.mock import ANY, Mock

import flask
import pytest
import werkzeug.exceptions
from cloudevents.abstract import CloudEvent
from functions_framework.execution_id import EXECUTION_ID_REQUEST_HEADER
from pydantic_core import PydanticSerializationError
from typing_extensions import assert_never

from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    CompletedExecutionResult,
    ExecutionInfo,
)
from ucam_faas.handler_execution import (
    DEFAULT_EXECUTION_RESULT_LOGGER,
    DEFAULT_HTTP_RESPONSE_CREATOR,
    CloudEventFunctionsFrameworkHandlerExecutor,
    CloudEventFunctionsFrameworkHandlerExecutorFn,
    HttpResponseCreator,
    JsonRepresentation,
    StructlogJsonExecutionResultLogger,
    _CloudEventHandlerT,
    cloud_event_handler_executor,
    get_function_execution_id,
)
from ucam_faas.tests.fixtures import make_cloud_event


def get_response_text(response: flask.Response) -> str:
    return "".join(
        chunk.decode() if isinstance(chunk, bytes) else chunk for chunk in response.response
    )


class TestHttpResponseCreator:
    RESPONSES: Final = {
        **DEFAULT_HTTP_RESPONSE_CREATOR.responses,
        "custom.abort": HttpResponseCreator.ResponseFields(
            status_code=http.client.IM_A_TEAPOT, message="I'm a teapot"
        ),
    }

    @pytest.fixture
    def http_response_creator(self) -> HttpResponseCreator:
        return HttpResponseCreator(responses=TestHttpResponseCreator.RESPONSES)

    @pytest.fixture(params=list(RESPONSES.items()), ids=lambda item: item[0])
    def aborted_response(
        self, request: pytest.FixtureRequest
    ) -> tuple[AbortedExecutionReason | str, HttpResponseCreator.ResponseFields]:
        reason, fields = request.param

        assert isinstance(reason, (AbortedExecutionReason, str))
        assert isinstance(fields, HttpResponseCreator.ResponseFields)

        return reason, fields

    def test_returns_none_for_execution_status_completed(
        self, http_response_creator: HttpResponseCreator
    ) -> None:
        # We return None on success because functions_framework treats handlers
        # returning None as successful and it generates its own success response.
        assert http_response_creator(ExecutionInfo(execution=CompletedExecutionResult())) is None

    def test_returns_aborted_response_for_configured_responses(
        self,
        http_response_creator: HttpResponseCreator,
        aborted_response: tuple[AbortedExecutionReason | str, HttpResponseCreator.ResponseFields],
    ) -> None:
        reason, fields = aborted_response

        resp = http_response_creator(ExecutionInfo(execution=AbortedExecutionResult(reason)))
        assert isinstance(resp, flask.Response)
        assert resp.status_code == fields.status_code
        assert resp.content_type == "text/plain"
        assert get_response_text(resp) == fields.message

    def test_returns_internal_server_error_for_unknown_aborted_reason(
        self, http_response_creator: HttpResponseCreator
    ) -> None:
        resp = http_response_creator(
            ExecutionInfo(execution=(AbortedExecutionResult(reason="unknown.foo")))
        )
        assert resp is not None
        assert resp.status_code == http.client.INTERNAL_SERVER_ERROR
        assert resp.content_type == "text/plain"
        assert get_response_text(resp) == "Failed to handle event for unknown reason"


def test_DEFAULT_HTTP_RESPONSE_CREATOR_defines_common_responses() -> None:
    assert DEFAULT_HTTP_RESPONSE_CREATOR.responses.keys() == set(AbortedExecutionReason)


class TestStructlogJsonExecutionResultLogger:
    @pytest.fixture
    def result_logger(
        self, json_representation: JsonRepresentation
    ) -> StructlogJsonExecutionResultLogger:
        return StructlogJsonExecutionResultLogger(json_representation=json_representation)

    @pytest.fixture(params=JsonRepresentation)
    def json_representation(self, request: pytest.FixtureRequest) -> JsonRepresentation:
        return JsonRepresentation(request.param)

    def test_logs_execution_info_for_completed_execution(
        self,
        caplog: pytest.LogCaptureFixture,
        result_logger: StructlogJsonExecutionResultLogger,
        json_representation: JsonRepresentation,
    ) -> None:
        result_logger(
            ExecutionInfo(
                execution=CompletedExecutionResult(duration_seconds=0.123),
                foo=123,
                bar={"baz": "boz"},
            )
        )

        assert len(caplog.records) == 1
        (record,) = caplog.records

        assert record.name == "ucam_faas"
        assert record.levelname == "INFO"
        msg = record.msg
        assert isinstance(msg, dict)

        if json_representation == JsonRepresentation.WIDE_FLAT:
            expected_fields = {
                "event": "ucam_faas_function_completed",
                "execution.status": "COMPLETED",
                "execution.duration_seconds": 0.123,
                "foo": 123,
                "bar.baz": "boz",
            }
        elif json_representation == JsonRepresentation.NARROW_DEEP:
            expected_fields = {
                "event": "ucam_faas_function_completed",
                "execution": {"duration_seconds": 0.123, "status": "COMPLETED"},
                "foo": 123,
                "bar": {"baz": "boz"},
            }
        else:
            assert_never(json_representation)

        assert {field: msg[field] for field in expected_fields} == expected_fields

    def test_logs_execution_info_for_aborted_execution(
        self,
        caplog: pytest.LogCaptureFixture,
        result_logger: StructlogJsonExecutionResultLogger,
        json_representation: JsonRepresentation,
    ) -> None:
        try:
            raise ValueError("Not OK")
        except ValueError as e:
            err = e

        result_logger(
            ExecutionInfo(
                execution=AbortedExecutionResult(
                    reason=AbortedExecutionReason.USER_DATA_INVALID,
                    exception=err,
                    duration_seconds=0.123,
                ),
                foo=123,
                bar={"baz": "boz"},
            )
        )

        assert len(caplog.records) == 1
        (record,) = caplog.records

        assert record.name == "ucam_faas"
        assert record.levelname == "ERROR"
        msg = record.msg
        assert isinstance(msg, dict)

        if json_representation == JsonRepresentation.WIDE_FLAT:
            expected_fields = {
                "event": "ucam_faas_function_aborted",
                "execution.status": "ABORTED",
                "execution.reason": "ucam_faas.user-data-invalid",
                "execution.duration_seconds": 0.123,
                "execution.exception": ANY,
                "foo": 123,
                "bar.baz": "boz",
            }
            msg_exception = msg["execution.exception"]
        elif json_representation == JsonRepresentation.NARROW_DEEP:
            expected_fields = {
                "event": "ucam_faas_function_aborted",
                "execution": {
                    "status": "ABORTED",
                    "reason": "ucam_faas.user-data-invalid",
                    "duration_seconds": 0.123,
                    "exception": ANY,
                },
                "foo": 123,
                "bar": {"baz": "boz"},
            }
            msg_exception = msg["execution"]["exception"]
        else:
            assert_never(json_representation)

        assert {field: msg[field] for field in expected_fields} == expected_fields

        assert msg_exception.startswith("Traceback (most recent call last):")
        assert "in test_logs_execution_info_for_aborted_execution" in msg_exception
        assert "ValueError: Not OK" in msg_exception

    def test_logs_exception_when_execution_info_json_serialization_fails(
        self,
        caplog: pytest.LogCaptureFixture,
        result_logger: StructlogJsonExecutionResultLogger,
        json_representation: JsonRepresentation,
    ) -> None:
        with pytest.raises(
            PydanticSerializationError, match=r"Unable to serialize unknown type: <class 'object'>"
        ):
            result_logger(ExecutionInfo(custom=object()))  # not JSON-serialisable

        assert len(caplog.records) == 1
        (record,) = caplog.records

        assert record.name == "ucam_faas"
        assert record.levelname == "ERROR"
        msg = record.msg
        assert isinstance(msg, dict)

        assert msg["event"] == "execution_info_json_serialisation_failed"
        assert (
            "PydanticSerializationError: Unable to serialize unknown type: <class 'object'>"
        ) in msg["exception"]


def test_DEFAULT_EXECUTION_RESULT_LOGGER() -> None:
    assert isinstance(DEFAULT_EXECUTION_RESULT_LOGGER, StructlogJsonExecutionResultLogger)
    assert DEFAULT_EXECUTION_RESULT_LOGGER.json_representation == JsonRepresentation.NARROW_DEEP


class Test_cloud_event_handler_executor:
    class GetDecoratedHandlerFn(Protocol):
        def __call__(
            self, handler: _CloudEventHandlerT
        ) -> CloudEventFunctionsFrameworkHandlerExecutorFn[_CloudEventHandlerT]:
            ...

    @pytest.fixture(autouse=True)
    def mock_ff_cloud_event(self) -> Generator[Mock]:
        with mock.patch("functions_framework.cloud_event") as mock_fn:
            yield mock_fn

    @pytest.fixture
    def log_execution_result(self) -> Mock:
        return Mock(name="log_execution_result")

    @pytest.fixture
    def create_http_response(self) -> Mock:
        return Mock(name="create_http_response", return_value=None)

    @staticmethod
    def raw_handler(cloud_event: CloudEvent, /) -> ExecutionInfo:
        return ExecutionInfo()

    @pytest.fixture
    def decorate_handler(
        self, log_execution_result: Mock, create_http_response: Mock
    ) -> GetDecoratedHandlerFn:
        def decorate_handler(
            handler: _CloudEventHandlerT,
        ) -> CloudEventFunctionsFrameworkHandlerExecutorFn[_CloudEventHandlerT]:
            # Equivalent to:
            # @cloud_event_handler_executor(
            #     log_execution_result=log_execution_result,
            #     create_http_response=create_http_response,
            # )
            # def handler(...):
            #     ...

            return cloud_event_handler_executor(
                log_execution_result=log_execution_result,
                create_http_response=create_http_response,
            )(handler)

        return decorate_handler

    @pytest.fixture
    def cloudevent(self) -> CloudEvent:
        return make_cloud_event(
            type="com.example.sampletype1", source="https://example.com/event-producer"
        )

    def test_is_Function_instance(self, decorate_handler: GetDecoratedHandlerFn) -> None:
        # functions_framework fails if handler functions are not instances of
        # types.FunctionType. (Ours is a callable class, so we have to fake this...)

        assert isinstance(decorate_handler(self.raw_handler), types.FunctionType)

    def test_has_name_of_wrapped_handler_fn(self, decorate_handler: GetDecoratedHandlerFn) -> None:
        # functions_framework uses the __name__ of handler functions to identify
        # them, so the name must be preserved when wrapping a user-provided
        # function.
        assert self.raw_handler.__name__ == "raw_handler"
        assert decorate_handler(self.raw_handler).__name__ == "raw_handler"

    def test_registers_handler_with_functions_framework(
        self, mock_ff_cloud_event: Mock, decorate_handler: GetDecoratedHandlerFn
    ) -> None:
        mock_ff_cloud_event.assert_not_called()

        wrapped_handler = decorate_handler(self.raw_handler)

        mock_ff_cloud_event.assert_called_once_with(wrapped_handler)

    def test_handler_fn_is_wrapped_as_CloudEventFunctionsFrameworkHandlerExecutor(
        self, decorate_handler: GetDecoratedHandlerFn
    ) -> None:
        wrapped_handler = decorate_handler(self.raw_handler)

        assert wrapped_handler is not self.raw_handler  # type: ignore[comparison-overlap]
        assert isinstance(wrapped_handler, CloudEventFunctionsFrameworkHandlerExecutor)

    def test_raw_handler_is_property_of_wrapped_handler(
        self, decorate_handler: GetDecoratedHandlerFn
    ) -> None:
        wrapped_handler = decorate_handler(self.raw_handler)
        assert wrapped_handler.execute is self.raw_handler

    @pytest.mark.parametrize("execute_for_result", [True, False])
    def test_logs_execution_result(
        self,
        decorate_handler: GetDecoratedHandlerFn,
        log_execution_result: Mock,
        cloudevent: CloudEvent,
        execute_for_result: bool,
    ) -> None:
        execution_info = ExecutionInfo(foo=123)

        def raw_handler(handler_cloudevent: CloudEvent, /) -> ExecutionInfo:
            assert handler_cloudevent is cloudevent
            return execution_info

        wrapped_handler = decorate_handler(raw_handler)
        if execute_for_result:
            wrapped_handler.execute_for_result(cloudevent)
        else:
            wrapped_handler(cloudevent)

        log_execution_result.assert_called_once_with(execution_info)

    def test_sends_response_from_create_http_response_fn(
        self,
        decorate_handler: GetDecoratedHandlerFn,
        create_http_response: Mock,
        cloudevent: CloudEvent,
    ) -> None:
        execution_info = ExecutionInfo(foo=123)
        expected_response = flask.Response("Example", status=http.client.IM_A_TEAPOT)
        create_http_response.return_value = expected_response

        def raw_handler(handler_cloudevent: CloudEvent, /) -> ExecutionInfo:
            assert handler_cloudevent is cloudevent
            return execution_info

        wrapped_handler = decorate_handler(raw_handler)

        # @functions_framework.cloud_event() handler does not allow the handler
        # to **return** a response, so we send responses by throwing HTTPException
        # which Flask/werkzeug handle.
        with pytest.raises(werkzeug.exceptions.HTTPException) as exc_info:
            wrapped_handler(cloudevent)

        assert exc_info.value.response is expected_response
        create_http_response.assert_called_once_with(execution_info)

    def test_does_not_send_response_when_create_http_response_fn_returns_none(
        self,
        decorate_handler: GetDecoratedHandlerFn,
        create_http_response: Mock,
        cloudevent: CloudEvent,
    ) -> None:
        execution_info = ExecutionInfo(foo=123)
        create_http_response.return_value = None

        def raw_handler(handler_cloudevent: CloudEvent, /) -> ExecutionInfo:
            assert handler_cloudevent is cloudevent
            return execution_info

        wrapped_handler = decorate_handler(raw_handler)

        assert wrapped_handler(cloudevent) is None  # type: ignore[func-returns-value]

        create_http_response.assert_called_once_with(execution_info)


def test_get_function_execution_id() -> None:
    execution_id = "Abcd1234Efgh"

    assert get_function_execution_id() is None

    app = flask.Flask("__future__")
    with app.test_request_context(headers={EXECUTION_ID_REQUEST_HEADER: execution_id}):
        assert flask.has_request_context()

        assert get_function_execution_id() == execution_id

    # Outside a flask context, IDs are unset
    assert get_function_execution_id() is None



from __future__ import annotations

import json
from datetime import datetime
from typing import TYPE_CHECKING, Iterable, Literal, Mapping, TypedDict, overload

from cloudevents.abstract import AnyCloudEvent, CloudEvent
from cloudevents.conversion import from_dict
from cloudevents.pydantic.v2 import CloudEvent as PydanticCloudEvent
from google.protobuf.message import Message
from pydantic import BaseModel
from typing_extensions import Required, Unpack

from ucam_faas.gcp_pubsub import (
    PUBSUB_CLOUD_EVENT_TOPIC_PATTERN,
    PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE,
    MessagePublishedData,
    PubsubMessage,
)

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem


def pubsub_message_bytes(cloud_event: CloudEvent) -> bytes:
    return MessagePublishedData.model_validate(cloud_event.get_data()).message.data


def make_pubsub_message(
    *,
    data: BaseModel | Mapping[str, object] | Message | bytes | None = None,
    attributes: Mapping[str, str] | None = None,
) -> PubsubMessage:
    _data: bytes
    if isinstance(data, BaseModel):
        _data = data.model_dump_json().encode()
    elif isinstance(data, Mapping):
        _data = json.dumps(data).encode()
    elif isinstance(data, Message):
        _data = data.SerializeToString()
    elif data is None:
        _data = b"__example__"
    elif isinstance(data, bytes):
        _data = data
    else:
        raise TypeError(f"Unsupported data type: {data!r}")
    return PubsubMessage(
        data=_data,
        attributes={**(attributes or {})},
        message_id="1",
        publish_time="2025-03-18T09:01:02.345Z",
    )


def make_pubsub_cloud_event(
    pubsub_message: PubsubMessage, topic: str | None = "projects/my-project/topics/my-topic"
) -> CloudEvent:
    if topic and not PUBSUB_CLOUD_EVENT_TOPIC_PATTERN.match(topic):
        raise ValueError(f"Invalid topic value: {topic}")

    ce_data_json = MessagePublishedData(message=pubsub_message).model_dump(
        mode="json", by_alias=True
    )
    return make_cloud_event(
        id=ce_data_json["message"]["messageId"],
        time=ce_data_json["message"]["publishTime"],
        specversion="1.0",
        datacontenttype="application/json",
        type=PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE,
        # functions_framework (which creates this value at runtime, when
        # converting the Pubsub-specific event data to CloudEvent format)
        # will use an empty string when the request URL invoking the
        # function does not match the pattern it expects.
        source=f"//pubsub.googleapis.com/{topic or ''}",
        data=MessagePublishedData(message=pubsub_message).model_dump(mode="json", by_alias=True),
    )


class CloudEventCreateAttributes(TypedDict, total=False):
    """
    Attributes required to create CloudEvent with from_dict().

    Attributes are `Required` if they have no default value. For example, `id`
    is required in CloudEvent objects, but it defaults to a generated UUID.

    See: https://github.com/cloudevents/spec/blob/v1.0.2/cloudevents/spec.md#context-attributes
    """

    id: str | None
    source: Required[str]
    specversion: Literal["1.0"]
    type: Required[str]
    datacontenttype: str
    dataschema: str
    subject: str
    time: int | str | datetime | None
    data: object


@overload
def make_cloud_event(
    event_type: type[PydanticCloudEvent] = ...,
    extra_attributes: SupportsKeysAndGetItem[str, object]
    | Iterable[tuple[str, object]]
    | None = ...,
    **attributes: Unpack[CloudEventCreateAttributes],
) -> PydanticCloudEvent:
    ...


@overload
def make_cloud_event(
    event_type: type[AnyCloudEvent],
    extra_attributes: SupportsKeysAndGetItem[str, object]
    | Iterable[tuple[str, object]]
    | None = ...,
    **attributes: Unpack[CloudEventCreateAttributes],
) -> AnyCloudEvent:
    ...


def make_cloud_event(
    event_type: type[AnyCloudEvent] | type[PydanticCloudEvent] = PydanticCloudEvent,
    extra_attributes: SupportsKeysAndGetItem[str, object]
    | Iterable[tuple[str, object]]
    | None = None,
    **attributes: Unpack[CloudEventCreateAttributes],
) -> AnyCloudEvent | PydanticCloudEvent:
    """Create a CloudEvent object with provided attribute values.

    `type` and `source` attributes are required, `id`, `specversion`, `time` and
    `data` have default values.

    See: https://github.com/cloudevents/spec/blob/v1.0.2/cloudevents/spec.md#context-attributes

    Examples
    --------
    >>> ce = make_cloud_event(
    ...     id='00000000-0000-0000-0000-000000000000',
    ...     type="com.example.sampletype1",
    ...     source="https://example.com/event-producer",
    ...     time=0,
    ...     data="Hi",
    ... )
    >>> type(ce)
    <class 'cloudevents.pydantic.v2.event.CloudEvent'>

    >>> from cloudevents.conversion import to_json
    >>> print(to_json(ce).decode())
    {\"specversion": "1.0", \
"id": "00000000-0000-0000-0000-000000000000", \
"source": "https://example.com/event-producer", \
"type": "com.example.sampletype1", \
"time": "1970-01-01T00:00:00+00:00", \
"data": "Hi"}
    """

    return from_dict(
        event_type=event_type,
        event={**(dict(extra_attributes) if extra_attributes is not None else {}), **attributes},
    )



import pytest

from ucam_faas.exceptions import UCAMFAASCouldNotLoadTarget
from ucam_faas.loading import load_function

MY_DICTIONARY: dict[str, str] = {}


def my_function() -> str:
    return "function_called"


def test_loading() -> None:
    func = load_function("my_function", "ucam_faas/tests/test_loading.py")

    assert func() == "function_called"


def test_loading_file_not_valid_module() -> None:
    with pytest.raises(UCAMFAASCouldNotLoadTarget):
        _ = load_function("my_function", "ucam_faas/tests/not_a_module.txt")


def test_loading_module_doesnt_contain_target() -> None:
    with pytest.raises(UCAMFAASCouldNotLoadTarget):
        _ = load_function("not_my_function", "ucam_faas/tests/test_loading.py")


def test_loading_target_isnt_a_function() -> None:
    with pytest.raises(UCAMFAASCouldNotLoadTarget):
        _ = load_function("MY_DICTIONARY", "ucam_faas/tests/test_loading.py")



from __future__ import annotations

from contextvars import ContextVar
from typing import Callable, Optional

import pytest
from typing_extensions import assert_type

from ucam_faas.contexts import (
    ActivateContextVarFn,
    contextual_default_factory,
    contextvar_activator,
)

EXAMPLE_CONTEXT: ContextVar[int] = ContextVar("EXAMPLE_CONTEXT")


def test_contextvar_activator() -> None:
    activate_example_context = contextvar_activator(EXAMPLE_CONTEXT)

    @activate_example_context(value=100)
    def func_with_context_active() -> None:
        assert EXAMPLE_CONTEXT.get() == 100

    # The context holds no value yet
    with pytest.raises(LookupError):
        assert EXAMPLE_CONTEXT.get()

    with activate_example_context(value=42) as activate_context_value:
        assert activate_context_value == 42
        assert EXAMPLE_CONTEXT.get() == 42

        # Contexts can be nested
        with activate_example_context(value=3) as inner_context_value:
            assert inner_context_value == 3
            assert EXAMPLE_CONTEXT.get() == 3

            func_with_context_active()

        # The old value is restored after the inner block ends
        assert EXAMPLE_CONTEXT.get() == 42

    # The context holds no value after execution leaves the with block.
    with pytest.raises(LookupError):
        assert EXAMPLE_CONTEXT.get()


class Test_contextual_default_factory:
    @pytest.fixture(scope="class")
    def activate_example_context(self) -> ActivateContextVarFn[int]:
        return contextvar_activator(EXAMPLE_CONTEXT)

    def test_overload_no_fn_no_default_fn(
        self, activate_example_context: ActivateContextVarFn[int]
    ) -> None:
        default_factory = contextual_default_factory(EXAMPLE_CONTEXT)

        assert_type(default_factory, Callable[[], int])

        with pytest.raises(LookupError):
            default_factory()

        with activate_example_context(42):
            assert default_factory() == 42

    def test_overload_no_fn(self, activate_example_context: ActivateContextVarFn[int]) -> None:
        default_factory = contextual_default_factory(EXAMPLE_CONTEXT, default_fn=lambda err: None)

        assert_type(default_factory, Callable[[], Optional[int]])

        assert default_factory() is None

        with activate_example_context(42):
            assert default_factory() == 42

    def test_overload_no_default_fn(
        self, activate_example_context: ActivateContextVarFn[int]
    ) -> None:
        default_factory = contextual_default_factory(EXAMPLE_CONTEXT, fn=lambda ctx: str(ctx))

        assert_type(default_factory, Callable[[], str])

        with pytest.raises(LookupError):
            default_factory()

        with activate_example_context(42):
            assert default_factory() == "42"

    def test_overload_fn_and_default_fn(
        self, activate_example_context: ActivateContextVarFn[int]
    ) -> None:
        default_factory = contextual_default_factory(
            EXAMPLE_CONTEXT, fn=lambda ctx: str(ctx), default_fn=lambda err: None
        )

        assert_type(default_factory, Callable[[], Optional[str]])

        assert default_factory() is None

        with activate_example_context(42):
            assert default_factory() == "42"



import os
import sys

import pytest
from cloudevents.pydantic.v2 import CloudEvent

from ucam_faas.testing import CloudEventFactory, cleanup_functions_framework


def test_cloudeventfactory_model_static_type_has_pydantic_methods() -> None:
    event = CloudEventFactory.build()
    assert CloudEvent.model_validate(event.model_dump()) == event

    # However type checkers don't see Pydantic methods on cloudevents.pydantic.CloudEvent
    from cloudevents.pydantic import CloudEvent as CloudEventAutoVersion

    # mypy complains: "type[CloudEvent]" has no attribute "model_validate"
    assert CloudEventAutoVersion.model_validate(event.model_dump()) == event


def test_cleanup_functions_framework(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.syspath_prepend("/bar")
    monkeypatch.syspath_prepend("/foo")
    syspath_before = list(sys.path)
    # syspath_before = list({p: None for p in sys.path}.keys())

    monkeypatch.delenv("FUNCTION_TARGET", raising=False)
    monkeypatch.setenv("FUNCTION_SIGNATURE_TYPE", "foo")
    monkeypatch.setenv("FUNCTION_TRIGGER_TYPE", "")

    with cleanup_functions_framework():
        sys.path.extend(["/foo", "/bar"] * 2)
        monkeypatch.setenv("FUNCTION_TARGET", "A")
        monkeypatch.setenv("FUNCTION_SIGNATURE_TYPE", "B")
        monkeypatch.setenv("FUNCTION_TRIGGER_TYPE", "C")

    assert syspath_before == sys.path

    assert "FUNCTION_TARGET" not in os.environ
    assert os.environ.get("FUNCTION_SIGNATURE_TYPE") == "foo"
    assert os.environ.get("FUNCTION_TRIGGER_TYPE") == ""



from __future__ import annotations

from typing import TypeVar

from ucam_faas.execution_info import ExecutionInfo


class UCAMFAASException(Exception):
    pass


class UCAMFAASCouldNotProcess(UCAMFAASException):
    pass


class UCAMFAASCouldNotLoadTarget(UCAMFAASException):
    pass


class ExecutionAbortedException(UCAMFAASException):
    execution_info: ExecutionInfo

    def __init__(self, *args: object, execution_info: ExecutionInfo) -> None:
        super().__init__(*args)
        self.execution_info = execution_info

    def __repr__(self) -> str:
        args_repr = f"{repr(self.args)[1:-1].rstrip(',')}, " if self.args else ""
        kwargs_repr = f"execution_info={self.execution_info!r}"
        return f"{self.__class__.__name__}({args_repr}{kwargs_repr})"


_ExceptionT = TypeVar("_ExceptionT", bound=BaseException)


def with_notes(exc: _ExceptionT, *notes: str) -> _ExceptionT:
    """
    Attach note strings to an Exception and return it.

    Versions of Python before 3.11 will not render notes in tracebacks, but will
    still have them attached in `__notes__`.
    """
    if not notes:
        return exc
    all_notes: list[str] = exc.__dict__.setdefault("__notes__", [])
    all_notes.extend(notes)
    return exc



"""
Support for event handler functions that process message events, such as GCP
PubSub messages.

This module provides support for:

- Validating incoming message metadata envelope
- Parsing/validating the message payload
- Recording the message envelope metadata in the function's `ExecutionInfo` result

Handler functions can focus on processing decoded message data payloads, without
needing to do their own parsing and validation or event metadata
logging/reporting.

To use this module, handler functions are wrapped with this module's
`@message_handler(...)` decorator. The decorator takes a `message_type`
argument that describes the message payload that the handler will be called
with to handle an incoming message. The `message_type` can be any of:

- A Pydantic model, to handle a JSON payload;
  the handler receives an instance of the model.
- A Protobuf Message class, to handle a protobuf payload;
  the handler receives an instance of the Message.
- `None` to perform no validation; the handler receives the payload bytes.
- A parser function that receives bytes and returns a value;
  the handler receives the parser's return value.

Handler functions receive the message payload decoded by the `message_type`, and
also have access to the the full GCP PubSub message and CloudEvent values that
the message was delivered with. Metadata from these metadata envelopes is
automatically included in `PubsubMessageExecutionInfo` values created within the
handler function's scope.
"""

from __future__ import annotations

import functools
from contextvars import ContextVar
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Protocol, cast, overload

from pydantic import BaseModel, Field
from typing_extensions import Never, TypeGuard, TypeVar

from ucam_faas.cloud_events import (
    CloudEventHandlerExecutionInfo,
    CloudEventHandlerExecutorFn,
    CloudEventType,
    cloud_event_handler,
)
from ucam_faas.contexts import contextual_default_factory, contextvar_activator
from ucam_faas.contextual_handlers import (
    ContextualHandlerParserFn,
    CreateExecutionInfoFn,
    HandlerFn,
    contextual_handler,
)
from ucam_faas.exceptions import ExecutionAbortedException, with_notes
from ucam_faas.execution_info import (
    AbortedExecutionReason,
    AbortedExecutionResult,
    ExecutionInfo,
)
from ucam_faas.gcp_pubsub import (
    PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE,
    MessagePublishedData,
    PubsubMessage,
)
from ucam_faas.handler_execution import HandlerExecutionOptions

if TYPE_CHECKING:
    # protobuf is not a required dependency
    from google.protobuf.message import Message

_PydanticModelT = TypeVar("_PydanticModelT", bound=BaseModel)
_ProtobufMessageT = TypeVar("_ProtobufMessageT", bound="Message")
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)

_InfoT = TypeVar("_InfoT", bound=ExecutionInfo)
_DefaultInfoT = TypeVar("_DefaultInfoT", bound=ExecutionInfo)
_DefaultInfoT_co = TypeVar("_DefaultInfoT_co", bound=ExecutionInfo, covariant=True)

_NoneT = TypeVar("_NoneT", None, Never, default=Never)


MESSAGE_CONTEXT: ContextVar[MessageHandlerContext] = ContextVar("MESSAGE_CONTEXT")
activate_message_context = contextvar_activator(MESSAGE_CONTEXT)
PUBSUB_MESSAGE_PUBLISHED_CLOUD_EVENT = CloudEventType(
    event_type=PUBSUB_MSG_PUBLISHED_CLOUD_EVENT_TYPE, event_data_model=MessagePublishedData
)


@dataclass
class MessageHandlerContext:
    pubsub_message: PubsubMessage
    handler: MessageHandlerAttributes


@dataclass
class MessageHandlerAttributes:
    input_message_accepted: bool


class PubsubMessageExecutionInfo(ExecutionInfo):
    pubsub_message: PubsubMessage = Field(
        default_factory=contextual_default_factory(
            MESSAGE_CONTEXT, fn=lambda ctx: ctx.pubsub_message
        )
    )
    message_handler: MessageHandlerAttributes = Field(
        default_factory=contextual_default_factory(MESSAGE_CONTEXT, fn=lambda ctx: ctx.handler)
    )


class MessageHandlerExecutionInfo(PubsubMessageExecutionInfo, CloudEventHandlerExecutionInfo):
    pass


class MessageParserFn(Protocol[_T_co]):
    def __call__(self, raw_message: bytes, /) -> _T_co:
        pass


class MessageHandlerDecoratorFn(Protocol[_T_co, _DefaultInfoT_co]):
    """The type of the decorator returned by `@message_handler(...)`"""

    # These overloads are needed to correctly type the return value of
    # the decorated handler function. The return varies according to whether the
    # handler function returns None:
    #
    # - When the handler CAN return None, the return is either the default
    #   ExecutionInfo type or the handler's non-None return type (if any).
    #   returned.
    # - When the handler NEVER returns None, the return is always the
    #   handler's own return type.

    @overload
    def __call__(
        self, event_data_handler: HandlerFn[_T_co, _InfoT], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT]:
        pass

    @overload
    def __call__(
        self, event_data_handler: HandlerFn[_T_co, _InfoT | None], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT | _DefaultInfoT_co]:
        pass

    def __call__(
        self, event_data_handler: HandlerFn[_T_co, _InfoT | _NoneT], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT | _DefaultInfoT_co]:
        pass


@overload
def message_handler(
    *,
    message_type: type[_PydanticModelT],
    create_default_info: None = None,
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_PydanticModelT, MessageHandlerExecutionInfo]:
    pass


@overload
def message_handler(
    *,
    message_type: type[_PydanticModelT],
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT],
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_PydanticModelT, _DefaultInfoT]:
    pass


@overload
def message_handler(
    *,
    message_type: type[_ProtobufMessageT],
    create_default_info: None = None,
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_ProtobufMessageT, MessageHandlerExecutionInfo]:
    pass


@overload
def message_handler(
    *,
    message_type: type[_ProtobufMessageT],
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT],
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_ProtobufMessageT, _DefaultInfoT]:
    pass


@overload
def message_handler(
    *,
    message_type: MessageParserFn[_T],
    create_default_info: None = None,
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_T, MessageHandlerExecutionInfo]:
    pass


@overload
def message_handler(
    *,
    message_type: MessageParserFn[_T],
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT],
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_T, _DefaultInfoT]:
    pass


@overload
def message_handler(
    *,
    message_type: None,
    create_default_info: None = None,
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[bytes, MessageHandlerExecutionInfo]:
    pass


@overload
def message_handler(
    *,
    message_type: None,
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT],
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[bytes, _DefaultInfoT]:
    pass


def message_handler(
    *,
    message_type: type[BaseModel] | type[Message] | MessageParserFn[object] | None,
    create_default_info: CreateExecutionInfoFn[_DefaultInfoT] | None = None,
    options: HandlerExecutionOptions | None = None,
) -> MessageHandlerDecoratorFn[_T, _DefaultInfoT]:
    """Register a function to be called when Pubsub publishes a message.

    This is a (python) function decorator that is used to define the entry point
    of a `ucam_faas` managed (cloud) function. The registered (python) function
    is called by `ucam_faas` each time a message is published to the Pubsub
    topic the deployed (cloud) function is subscribed to. (The topic and other
    subscription details are configured at during deployment, not when
    registering a python function with this decorator.)

    The registered function receives a validated message value, parsed from the
    data payload of the incoming Pubsub event, using the Pydantic model,
    Protobuf message or parser function provided to the `@message_handler(...)`
    decorator when registering.

    Parameters
    ----------
    message_type:
        The type to use to decode/validate the message payload, or function to
        parse the payload. The handler function is called with decoded message
        payload. If `None`, the handler receives the payload as `bytes`, without
        any decoding.
    create_default_info:
        A function used to create an ExecutionInfo value if the decorated
        handler function does not return one.
    options:
        Override the default behaviour for logging the `ExecutionInfo` result,
        and representing the `ExecutionInfo` result as an HTTP response.

    Example
    -------
    First define the data payload of the message you will handle:
    >>> from pydantic import BaseModel
    >>> class GreetingMessage(BaseModel):
    ...     greeting: str

    Second, define a function to handle the message. It'll be called automatically
    each time a PubSub message is received — as long as the message passes
    validation using your model.

    >>> from ucam_faas.messages import message_handler
    >>> @message_handler(message_type=GreetingMessage)
    ... def handle_greeting(message: GreetingMessage) -> None:
    ...     print(f"Received greeting: {message.greeting!r}")

    The handler function can be executed for testing purposes:
    >>> from ucam_faas.tests.fixtures import make_pubsub_cloud_event, make_pubsub_message
    >>> cloud_event = make_pubsub_cloud_event(
    ...     make_pubsub_message(data=GreetingMessage(greeting="Hello World!")),
    ...     topic="projects/my-project/topics/greetings",
    ... )
    >>> exc_info = handle_greeting.execute_for_result(cloud_event)
    Received greeting: 'Hello World!'
    """
    message_parser_fn = get_message_parser_fn(message_type)
    parser = contextual_message_parser(cast(MessageParserFn[_T], message_parser_fn))

    @overload
    def message_handler_decorator(
        message_handler: HandlerFn[_T, _InfoT], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT]:
        pass

    @overload
    def message_handler_decorator(
        message_handler: HandlerFn[_T, _InfoT | None], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT | _DefaultInfoT]:
        pass

    def message_handler_decorator(
        message_handler: HandlerFn[_T, _InfoT | _NoneT], /
    ) -> CloudEventHandlerExecutorFn[MessagePublishedData, _InfoT | _DefaultInfoT]:
        # When create_default_info is None or
        # CreateExecutionInfoFn[MessageExecutionInfo],
        # _DefaultInfoT IS PubsubMessageExecutionInfo because of the
        # @overloads, so this cast is safe.
        _create_default_info = cast(
            CreateExecutionInfoFn[_DefaultInfoT],
            create_default_info or MessageHandlerExecutionInfo.from_context,
        )

        @cloud_event_handler(
            event_data_parser=PUBSUB_MESSAGE_PUBLISHED_CLOUD_EVENT, options=options
        )
        @contextual_handler(
            parser=parser,
            context=activate_message_context,
            default_info_factory=_create_default_info,
        )
        @functools.wraps(message_handler)
        def message_handler_executor(event_data: _T, /) -> _InfoT | _NoneT:
            return message_handler(event_data)

        assert message_handler_executor.__name__ == message_handler_executor.__name__
        return message_handler_executor

    return message_handler_decorator


def bytes_message_parser(raw_message: bytes, /) -> bytes:
    return raw_message


def is_protobuf_message_cls(value: object) -> TypeGuard[type[Message]]:
    return isinstance(value, type) and callable(getattr(value, "ParseFromString", None))


# This is a class rather than fn to allow introspection of the model being used,
# rather than having it hidden in a function closure.
@dataclass
class PydanticMessageParser(Generic[_PydanticModelT]):
    model: type[_PydanticModelT]

    def __call__(self, raw_message: bytes, /) -> _PydanticModelT:
        return self.model.model_validate_json(raw_message)


@dataclass
class ProtobufMessageParser(Generic[_ProtobufMessageT]):
    message_type: type[_ProtobufMessageT]

    def __call__(self, raw_message: bytes, /) -> _ProtobufMessageT:
        return self.message_type.FromString(raw_message)


@overload
def get_message_parser_fn(
    message_type: type[_PydanticModelT],
) -> MessageParserFn[_PydanticModelT]:
    pass


@overload
def get_message_parser_fn(
    message_type: type[_ProtobufMessageT],
) -> MessageParserFn[_ProtobufMessageT]:
    pass


@overload
def get_message_parser_fn(
    message_type: MessageParserFn[_T],
) -> MessageParserFn[_T]:
    pass


@overload
def get_message_parser_fn(
    message_type: None,
) -> MessageParserFn[bytes]:
    pass


def get_message_parser_fn(
    message_type: type[BaseModel] | type[Message] | MessageParserFn[object] | None,
) -> MessageParserFn[object]:
    if message_type is None:
        return bytes_message_parser
    elif isinstance(message_type, type) and issubclass(message_type, BaseModel):
        return PydanticMessageParser(model=message_type)
    elif is_protobuf_message_cls(message_type):
        return ProtobufMessageParser(message_type=message_type)
    elif callable(message_type):
        return cast(MessageParserFn[object], message_type)
    raise TypeError(
        f"message_parser must be a Pydantic model class, a Protobuf Message "
        f"class, a function or None, got: {message_type!r}"
    )


def parse_pubsub_message_data(
    message_parser: MessageParserFn[_T], event_data: MessagePublishedData
) -> _T:
    try:
        return message_parser(event_data.message.data)
    except ExecutionAbortedException:
        raise
    except Exception as e:
        raise with_notes(
            ExecutionAbortedException(
                "message_parser failed to parse message",
                execution_info=MessageHandlerExecutionInfo(
                    execution=AbortedExecutionResult(
                        reason=AbortedExecutionReason.USER_DATA_INVALID, exception=e
                    )
                ),
            ),
            f"See this exception's cause for full details:\n{e}",
        ) from e


def contextual_message_parser(
    message_parser: MessageParserFn[_T_co],
) -> ContextualHandlerParserFn[MessagePublishedData, _T_co, MessageHandlerContext, ExecutionInfo]:
    """Wrap a (bytes) -> T function to be a @contextual_handler() parser.

    The T value returned by the function is the primary value and the context
    contains the GCP Pub/Sub `MessagePublished` event that the data came from.

    The function is run with `MESSAGE_CONTEXT` active.
    """

    @functools.wraps(message_parser)
    def parse_input(
        event_data: MessagePublishedData, /
    ) -> tuple[_T_co, MessageHandlerContext] | ExecutionInfo:
        context = MessageHandlerContext(
            pubsub_message=event_data.message,
            handler=MessageHandlerAttributes(input_message_accepted=False),
        )

        with activate_message_context(context):
            parse_result = parse_pubsub_message_data(message_parser, event_data)

        context.handler.input_message_accepted = True
        return parse_result, context

    return parse_input
